@@ -11,6 +11,7 @@ from ai.backend.common.asyncio import cancel_tasks
 from ai.backend.plugin.entrypoint import scan_entrypoints
 
 from ..etcd import AsyncEtcd
+from ..exception import ConfigurationError
 from ..logging_utils import BraceStyleAdapter
 
 log = BraceStyleAdapter(logging.getLogger(__name__))
@@ -95,6 +96,7 @@ class BasePluginContext(Generic[P]):
     local_config: Mapping[str, Any]
     plugins: Dict[str, P]
     plugin_group: ClassVar[str] = "backendai_XXX_v10"
+    allowlist: ClassVar[Optional[set[str]]] = None
     blocklist: ClassVar[Optional[set[str]]] = None
 
     _config_watchers: WeakSet[asyncio.Task]
@@ -116,16 +118,40 @@ class BasePluginContext(Generic[P]):
     def discover_plugins(
         cls,
         plugin_group: str,
+        allowlist: set[str] = None,
         blocklist: set[str] = None,
     ) -> Iterator[Tuple[str, Type[P]]]:
+        cls_allowlist = set() if cls.allowlist is None else cls.allowlist
+        arg_allowlist = set() if allowlist is None else allowlist
+        allowlist_enabled = allowlist is not None or cls.allowlist is not None
         cls_blocklist = set() if cls.blocklist is None else cls.blocklist
         arg_blocklist = set() if blocklist is None else blocklist
-        for entrypoint in scan_entrypoints(plugin_group, cls_blocklist | arg_blocklist):
+        for entrypoint in scan_entrypoints(
+            plugin_group,
+            allowlist=cls_allowlist | arg_allowlist if allowlist_enabled else None,
+            blocklist=cls_blocklist | arg_blocklist,
+        ):
             log.info("loading plugin (group:{}): {}", plugin_group, entrypoint.name)
             yield entrypoint.name, entrypoint.load()
 
-    async def init(self, context: Any = None) -> None:
-        scanned_plugins = self.discover_plugins(self.plugin_group)
+    async def init(
+        self,
+        context: Any = None,
+        allowlist: Optional[set] = None,
+        blocklist: Optional[set] = None,
+    ) -> None:
+        if allowlist is not None and blocklist is not None:
+            if union := allowlist & blocklist:
+                raise ConfigurationError(
+                    {
+                        "plugin.BasePluginContext": f"allowlist and blocklist has union value '{union}'"
+                    }
+                )
+        scanned_plugins = self.discover_plugins(
+            self.plugin_group,
+            allowlist=allowlist,
+            blocklist=blocklist,
+        )
         for plugin_name, plugin_entry in scanned_plugins:
             plugin_config = await self.etcd.get_prefix(
                 f"config/plugins/{self._group_key}/{plugin_name}/",
