@@ -8,14 +8,16 @@ import numbers
 import sys
 import uuid
 from abc import ABCMeta, abstractmethod
-from collections import UserDict, namedtuple
+from collections import UserDict, defaultdict, namedtuple
 from contextvars import ContextVar
 from decimal import Decimal
+from ipaddress import ip_address, ip_network
 from pathlib import PurePosixPath
 from typing import (
     TYPE_CHECKING,
     Any,
     Dict,
+    Generic,
     List,
     Literal,
     Mapping,
@@ -31,12 +33,13 @@ from typing import (
     overload,
 )
 
-import aioredis
-import aioredis.client
-import aioredis.sentinel
-import attr
+import attrs
+import redis.asyncio.sentinel
 import trafaret as t
 import typeguard
+from redis.asyncio import Redis
+
+from .exception import InvalidIpAddressValue
 
 __all__ = (
     "aobject",
@@ -57,6 +60,7 @@ __all__ = (
     "SlotName",
     "IntrinsicSlotNames",
     "ResourceSlot",
+    "ReadableCIDR",
     "HardwareMetadata",
     "MountPermission",
     "MountPermissionLiteral",
@@ -210,6 +214,27 @@ AccessKey = NewType("AccessKey", str)
 SecretKey = NewType("SecretKey", str)
 
 
+class AbstractPermission(str, enum.Enum):
+    """
+    Abstract enum type for permissions
+    """
+
+
+class VFolderHostPermission(AbstractPermission):
+    """
+    Atomic permissions for a virtual folder under a host given to a specific access key.
+    """
+
+    CREATE = "create-vfolder"
+    MODIFY = "modify-vfolder"  # rename, update-options
+    DELETE = "delete-vfolder"
+    MOUNT_IN_SESSION = "mount-in-session"
+    UPLOAD_FILE = "upload-file"
+    DOWNLOAD_FILE = "download-file"
+    INVITE_OTHERS = "invite-others"  # invite other user to user-type vfolder
+    SET_USER_PERM = "set-user-specific-permission"  # override permission of group-type vfolder
+
+
 class LogSeverity(str, enum.Enum):
     CRITICAL = "critical"
     ERROR = "error"
@@ -240,6 +265,7 @@ class ServicePortProtocols(str, enum.Enum):
     HTTP = "http"
     TCP = "tcp"
     PREOPEN = "preopen"
+    INTERNAL = "internal"
 
 
 class SessionTypes(str, enum.Enum):
@@ -258,6 +284,11 @@ class ClusterMode(str, enum.Enum):
     MULTI_NODE = "multi-node"
 
 
+class CommitStatus(str, enum.Enum):
+    READY = "ready"
+    ONGOING = "ongoing"
+
+
 class MovingStatValue(TypedDict):
     min: str
     max: str
@@ -331,6 +362,58 @@ class HostPortPair(namedtuple("HostPortPair", "host port")):
         return f"{self.host}:{self.port}"
 
 
+_Address = TypeVar("_Address", bound=Union[ipaddress.IPv4Network, ipaddress.IPv6Network])
+
+
+class ReadableCIDR(Generic[_Address]):
+    """
+    Convert wild-card based IP address into CIDR.
+
+    e.g)
+    192.10.*.* -> 192.10.0.0/16
+    """
+
+    _address: _Address | None
+
+    def __init__(self, address: str | None, is_network: bool = True) -> None:
+        self._is_network = is_network
+        self._address = self._convert_to_cidr(address) if address is not None else None
+
+    def _convert_to_cidr(self, value: str) -> _Address:
+        str_val = str(value)
+        if not self._is_network:
+            return cast(_Address, ip_address(str_val))
+        if "*" in str_val:
+            _ip, _, given_cidr = str_val.partition("/")
+            filtered = _ip.replace("*", "0")
+            if given_cidr:
+                return self._to_ip_network(f"{filtered}/{given_cidr}")
+            octets = _ip.split(".")
+            cidr = octets.index("*") * 8
+            return self._to_ip_network(f"{filtered}/{cidr}")
+        return self._to_ip_network(str_val)
+
+    @staticmethod
+    def _to_ip_network(val: str) -> _Address:
+        try:
+            return cast(_Address, ip_network(val))
+        except ValueError:
+            raise InvalidIpAddressValue
+
+    @property
+    def address(self) -> _Address | None:
+        return self._address
+
+    def __str__(self) -> str:
+        return str(self._address)
+
+    def __eq__(self, other: object) -> bool:
+        if other is self:
+            return True
+        assert isinstance(other, ReadableCIDR), "Only can compare ReadableCIDR objects."
+        return self.address == other.address
+
+
 class BinarySize(int):
     """
     A wrapper around Python integers to represent binary sizes for storage and
@@ -688,7 +771,7 @@ class JSONSerializableMixin(metaclass=ABCMeta):
         raise NotImplementedError
 
 
-@attr.define(slots=True)
+@attrs.define(slots=True)
 class VFolderMount(JSONSerializableMixin):
     name: str
     vfid: uuid.UUID
@@ -727,6 +810,35 @@ class VFolderMount(JSONSerializableMixin):
         )
 
 
+class VFolderHostPermissionMap(dict, JSONSerializableMixin):
+    def __or__(self, other: Any) -> VFolderHostPermissionMap:
+        if self is other:
+            return self
+        if not isinstance(other, dict):
+            raise ValueError(f"Invalid type. expected `dict` type, got {type(other)} type")
+        union_map: Dict[str, set] = defaultdict(set)
+        for host, perms in [*self.items(), *other.items()]:
+            try:
+                perm_list = [VFolderHostPermission(perm) for perm in perms]
+            except ValueError:
+                raise ValueError(f"Invalid type. Permissions of Host `{host}` are ({perms})")
+            union_map[host] |= set(perm_list)
+        return VFolderHostPermissionMap(union_map)
+
+    def to_json(self) -> dict[str, Any]:
+        return {host: [perm.value for perm in perms] for host, perms in self.items()}
+
+    @classmethod
+    def from_json(cls, obj: Mapping[str, Any]) -> JSONSerializableMixin:
+        return cls(**cls.as_trafaret().check(obj))
+
+    @classmethod
+    def as_trafaret(cls) -> t.Trafaret:
+        from . import validators as tx
+
+        return t.Dict(t.String, t.List(tx.Enum(VFolderHostPermission)))
+
+
 class ImageRegistry(TypedDict):
     name: str
     url: str
@@ -750,12 +862,16 @@ class ServicePort(TypedDict):
     host_ports: Sequence[Optional[int]]
 
 
+ClusterSSHPortMapping = NewType("ClusterSSHPortMapping", Mapping[str, Tuple[str, int]])
+
+
 class ClusterInfo(TypedDict):
     mode: ClusterMode
     size: int
     replicas: Mapping[str, int]  # per-role kernel counts
     network_name: Optional[str]
     ssh_keypair: Optional[ClusterSSHKeyPair]
+    cluster_ssh_port_mapping: Optional[ClusterSSHPortMapping]
 
 
 class ClusterSSHKeyPair(TypedDict):
@@ -764,7 +880,7 @@ class ClusterSSHKeyPair(TypedDict):
 
 
 class DeviceModelInfo(TypedDict):
-    device_id: DeviceId
+    device_id: DeviceId | str
     model_name: str
     data: Mapping[str, Any]
 
@@ -780,6 +896,8 @@ class KernelCreationResult(TypedDict):
     repl_out_port: int
     stdin_port: int  # legacy
     stdout_port: int  # legacy
+    scaling_group: str
+    agent_addr: str
 
 
 class KernelCreationConfig(TypedDict):
@@ -800,6 +918,9 @@ class KernelCreationConfig(TypedDict):
     startup_command: Optional[str]
     internal_data: Optional[Mapping[str, Any]]
     preopen_ports: List[int]
+    allocated_host_ports: List[int]
+    scaling_group: str
+    agent_addr: str
 
 
 class KernelEnqueueingConfig(TypedDict):
@@ -848,11 +969,11 @@ class EtcdRedisConfig(TypedDict, total=False):
     password: Optional[str]
 
 
-@attr.s(auto_attribs=True)
+@attrs.define(auto_attribs=True)
 class RedisConnectionInfo:
-    client: aioredis.Redis | aioredis.sentinel.Sentinel
+    client: Redis | redis.asyncio.sentinel.Sentinel
     service_name: Optional[str]
 
     async def close(self) -> None:
-        if isinstance(self.client, aioredis.Redis):
+        if isinstance(self.client, Redis):
             await self.client.close()
